;;;  Modisc 
;;;  Ranjan Srivastava
;;;  Last modified - 2/26/06

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Copyright (c) 2006, Ranjan Srivastava
;All rights reserved.

;Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

;Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
;Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
;Neither the name of the University of Connecticut nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
;THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defstruct model
  name
  parameters
  variables
  dof
  data)

(defun modisc ()
  (let ((data-file (capi:prompt-for-file "Enter input file:")))
    (if (eql data-file nil)
        nil
      (analyze-data (with-open-file (src data-file :direction :input)
        (do ((dat (read src nil 'eof)  (read src nil 'eof))
             (buffer nil (cons dat buffer)))
            ((eql dat 'eof) (reverse buffer))))))))

(defun partition-list (data)
  (let ((buffer nil)
        (sublist nil)
        (comment-flag nil))
    (dolist (val data)
      (cond 
       ((eql val 'begin-comment) (setf comment-flag t))
       ((eql val 'end-comment) (setf comment-flag nil)))
      (cond
       (comment-flag)
       ((or (eql val 'end-model)
            (eql val 'end-experiment))
        (progn
          (push (reverse sublist) buffer)
          (setf sublist nil)))
       ((not (or (eql val 'parameters)
                 (eql val 'dof)
                 (eql val 'variables)
                 (eql val 'end-comment)))
        (push val sublist))))
    (reverse buffer)))

(defun organize-experiment-info (data)
  (if (not (eql (car data) 'begin-experiment))
      nil
    (let* ((vars (car (cdr data)))
          (events (- (/ (- (length data) 2) (car (cdr data))) 1))
          (exptl-data (make-array (list (+ 1 events) vars)))
          (data-list (cdr (cdr data))))
      (do ((i 0 (+ i 1)))
          ((> i events))
        (do ((j 0 (+ j 1)))
            ((> j (- vars 1)))
          (setf (aref exptl-data i j) (pop data-list))))
      exptl-data)))

(defun organize-model (data)
  (if (not (eql (car data) 'begin-model))
      nil
    (let* ((mod-data (cdr data))
           (sim-data (cdr (cdr (cdr (cdr (cdr mod-data))))))
           (vars (car (cdr (cdr mod-data))))
           (events (- (/ (- (length mod-data) 3) vars) 1)))
    (setf tmp-model (make-model
                     :name (pop mod-data)
                     :parameters (pop mod-data)
                     :variables (pop mod-data)
                     :dof (pop mod-data)
                     :data (make-array (list (- vars 1)) :initial-element nil)))
    (do ((i 0 (+ i 1)))
        ((> i (- vars 2)))
      (setf (svref (model-data tmp-model) i) (make-hash-table))
      (setf (gethash 'var-name (svref (model-data tmp-model) i)) (pop sim-data)))
    (do ((i 0 (+ i 1)))
        ((> i (- events 1)))
      (do ((j 0 (+ j 1)))
          ((> j (- vars 1)))
        (if (eql j 0)
            (setf ind (pop sim-data))
          (setf (gethash ind (svref (model-data tmp-model) (- j 1))) (pop sim-data)))))
    tmp-model)))
                     

(defun analyze-data (raw-data)
  (let* ((data (partition-list raw-data))
         (tmp-stack nil)
         (index 1)
         (number-of-models (- (length data) 1))
         (mod-vec (make-array (+ 1 number-of-models) :initial-element nil)))
    (dolist (val data)
      (if (eql (car val) 'begin-experiment)
          (push (pop data) tmp-stack)
        (progn
          (setf (svref mod-vec index) (organize-model (pop data)))
          (setf index (+ index 1)))))
    (setf (svref mod-vec 0) (organize-experiment-info (pop tmp-stack)))
    (gen-output mod-vec (perform-calculations mod-vec number-of-models) number-of-models)))

(defun gen-output (data prob-share num-of-models)
    (terpri)
    (do ((index 1 (+ index 1)))
        ((> index num-of-models) (terpri))
      (format t "The probability share for model ~A is ~A.~%" 
              (model-name (svref data index)) (svref prob-share index))))

(defun perform-calculations (data modnum)
  (let* ((results (make-array (+ modnum 1) :initial-element nil)))
    (setf (svref results 0) 0)
    (do ((i 1 (+ i 1)))
        ((> i modnum))
      (setf (svref results i) (calculate-probability (svref data 0) (svref data i)))
      (setf (svref results 0) (+ (svref results 0) (svref results i))))
    (do ((i 1 (+ i 1)))
        ((> i modnum))
      (setf (svref results i) (/ (svref results i) (svref results 0))))
    results))

(defun calculate-probability (exptl-dat mod-dat)
  (let* ((p (model-parameters mod-dat))
         (vars (model-variables mod-dat))
         (ve (model-dof mod-dat)))
    (* (expt 2 (/ (* -1 p) 2)) (expt (determinant (modified-covar exptl-dat mod-dat vars)) (/ (* -1 ve) 2d0)))))

(defun modified-covar (exptl sim vars)
  (let* ((results-matrix (make-array (list (- vars 1) (- vars 1)) :initial-element nil))
         (var-vector (make-array (list (- vars 1)) :initial-element nil))
         (n (car (array-dimensions exptl)))
         (mod-data (make-array (list n vars) :initial-element nil)))
    (do ((index 0 (+ index 1)))
        ((> index (- vars 2)))
      (setf (svref var-vector index) (get-data 'var-name sim index)))
    (setf mod-data (mod-array exptl var-vector))
    (do ((i 0 (+ i 1)))
        ((> i (- vars 2)))
      (do ((k 0 (+ k 1)))
          ((> k (- vars 2)))
        (setf (aref results-matrix i k) 
              (do ((u 1 (+ 1 u)) 
                   (val 0 (+ val (* (- (aref mod-data u i) (get-data (aref mod-data u (- vars 1)) sim i))
                                    (- (aref mod-data u k) (get-data (aref mod-data u (- vars 1)) sim k))))))
                  ((> u (- n 1)) val)))))
    results-matrix))
        
    
    
(defun get-data (ind-var model-id var-index)
  (gethash ind-var (svref (model-data model-id) var-index)))

(defun mod-array (expt-array var-vec)
  (let* ((rows (car (array-dimensions expt-array)))
         (cols (+ 1 (car (array-dimensions var-vec))))
         (index 0)
         (new-array (make-array (list rows cols) :initial-element nil)))
    (do ((i 0 (+ i 1)))
        ((> i (- rows 1)))
         (setf (aref new-array i (- cols 1)) (aref expt-array i 0)))
    (do ((j 0 (+ j 1)))
        ((> j (- (car (cdr (array-dimensions expt-array))) 1)))
      (if (not (position (aref expt-array 0 j) var-vec))
          nil
        (do ((i 0 (+ 1 i)))
            ((> i (- rows 1)) (setf index (+ 1 index)))
          (setf (aref new-array i index) (aref expt-array i j)))))
    new-array))
  

(defun determinant (mat)
  (let* ((n (car (array-dimensions mat)))
         (number-of-swaps 0)
         (tmp-vec (make-array n :initial-element nil)))
    (if (not (eql n (cadr (array-dimensions mat))))
        (format t "Not a square matrix")
      (dotimes (k (- n 1))
        (if (eql 0 (aref mat k k))
            (progn
              (setf number-of-swaps (+ 1 number-of-swaps))
              (swap-rows mat k n tmp-vec)))
        (make-lower-triangle mat k n)))
    (do ((k 0 (+ 1 k))
         (prod 1 (* prod (aref mat k k))))
        ((eql k n) 
         (if (oddp number-of-swaps)
             (* -1 prod)
           prod)))))

       
(defun swap-rows (mat k n tmp-vec)
  (do ((j (+ k 1) (+ j 1)))
      ((not (eql (aref mat j k) 0))
       (dotimes (i n)
         (setf (svref tmp-vec i) (aref mat j i))
         (setf (aref mat j i) (aref mat k i))
         (setf (aref mat k i) (svref tmp-vec i))))))

(defun make-lower-triangle (mat k n)
  (do ((j (+ k 1) (+ j 1)))
      ((> j (- n 1)))
    (let ((p (- (/ (aref mat j k) (aref mat k k)))))
      (do ((l k (+ l 1)))
          ((> l (- n 1)))
        (setf (aref mat j l) (+ (* p (aref mat k l)) (aref mat j l)))))))
    
          
  
  
         


      
      
      


 
          
          




      
       

        
        


      
        
            










     



        

